/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                       NESCarts.h                        **/
/**                                                         **/
/** This file contains definitions of different fields in   **/
/** the .NES file header. Also see NESCarts.c.              **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1998-2001                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#ifndef NESCARTS_H
#define NESCARTS_H

#define NES_ROMBanks(Header)  Header[4]
#define NES_VROMBanks(Header) Header[5]
#define NES_ROMSize(Header)   ((int)NES_ROMBanks(Header)<<14)
#define NES_VROMSize(Header)  ((int)NES_VROMBanks(Header)<<13)
#define NES_Mapper(Header)    ((Header[6]>>4)|(Header[7]&0xF0))
#define NES_Mirroring(Header) (Header[6]&0x01)
#define NES_Battery(Header)   (Header[6]&0x02)
#define NES_Trainer(Header)   (Header[6]&0x04)
#define NES_4Screens(Header)  (Header[6]&0x08)
#define NES_VSSystem(Header)  (Header[7]&0x01)
#define NES_RAMBanks(Header)  Header[8]
/** NES_CRC() ************************************************/
/** Calculate CRC of a given number of bytes.               **/
/*************************************************************/
unsigned short NES_CRC(unsigned char *Data,int N);

#define MAX_BANKS 10

typedef struct NesCart_
{
	HANDLE Hdl;
	short nbROMBanks, nbVROMBanks; /* Nomber of ROM Bank and VRom bank */
	long ROMSize, VROMSize;    /* Size of one ROM bank and one VRom bank */
	short nbRAMBanks;		
	char Mapper;			   /* Mapper Type */
	struct
	{
		unsigned char Mirroring:1;	/* 0> Horizontal mirroring : 1> Vertical mirroring */
		unsigned char Battery:1;	/* Battery save ?*/
		unsigned char Trainer:1;    /* A trainer in ?*/
		unsigned char FourScreens:1;/* Four screen ?*/
		unsigned char VSSystem:1;   /* ?? */
	} flags;
	unsigned char *dotNESfile; /* Pointer to the file in memory */
	unsigned char *Trainer;    /* Pointer to the start of the trainer */
	unsigned char *ROMBanks[MAX_BANKS]; /* Pointeur to the first ROM bank*/
	unsigned char *VROMBanks[MAX_BANKS]; /* Pointeur to the first ROM bank*/
} NesCart;

int LoadCart(const char *filename, NesCart *cart);
void ShowCartInfo(NesCart *card);

void UnloadCart(NesCart *cart);

#endif /* NESCARTS_H */