/* 
 * TI-NESulator project : main.c
 *
 *
 * Code Copyright  2003 by TRAPIER Manoel
 *
 */

#define USE_TI89              // Compile for TI-89
#define USE_TI92PLUS          // Compile for TI-92 Plus
#define USE_V200              // Compile for V200

// #define OPTIMIZE_ROM_CALLS // Use ROM Call Optimization

#define MIN_AMS 100           // Compile for AMS 1.00 or higher

//#define SAVE_SCREEN           // Save/Restore LCD Contents

#include <tigcclib.h>         // Include All Header File

#include "M6502.h"
#include "ppu.h"
#include "NESCarts.h"
#include "Paddle.h"

#include "Mappers.h"

/* We use Xlib */
#include "X2D.h"

unsigned char *Memory;
unsigned char *mLBank, *mUBank;

unsigned char FenX,FenY;

void *Buffer;

PPU ppu;

#define fatal(s) { DlgMessage(s,"TI-NESulator",BT_NONE,BT_OK); ST_showHelp("TI-NESulator by Godzil");	exit(0); }

Paddle P1;

unsigned short ScanLine;

SCR_RECT Scr;

void _main()
{
	NesCart Cart;
	M6502 MainCPU;
	LCD_BUFFER SavedScreen;
	
/* Here we will fill the memory */
/*
======================================= $10000
 Upper Bank of Cartridge ROM
--------------------------------------- $F000

--------------------------------------- $E000

--------------------------------------- $D000

======================================= $C000
 Lower Bank of Cartridge ROM
--------------------------------------- $B000

--------------------------------------- $A000

--------------------------------------- $9000

======================================= $8000
 Cartridge RAM (may be battery-backed)
--------------------------------------- $7000
 
======================================= $6000
 Expansion Modules
======================================= $5000
 Input/Output
--------------------------------------- $4000

--------------------------------------- $3000

======================================= $2000
 2kB Internal RAM, mirrored 4 times
--------------------------------------- $1000

--------------------------------------- $0000
*/
	LCD_save(SavedScreen);
	clrscr();
	FontSetSys(F_4x6);
	
	/* Ci dessous l'instruction la plus importante de l'mulateur =) */
	printf("TI-NESulator v%d.%d by Godzil\nCopyright 2003 TRAPIER Manol\n",0,2);

	Scr.l = 0;
	if (TI89)
	{
		Scr.xy.x1 = 160;
		Scr.xy.y1 = 100;		
	}
	else
	{ /* Screen for V200/92+ is same size */
		Scr.xy.x1 = 240;
		Scr.xy.y1 = 128;		
	}	

	
	printf("Please Wait while loading %s cartridge...\n","smb1.nes");
	if(LoadCart("smb1",&Cart) != 0)
		fatal("Loading error...\n");
	printf("Creating NES memory\n");
	printf("->Non ROM memory (32kB)      ");
	if ( (Memory = (unsigned char *)malloc(0x8000)) == NULL)
	{
		UnloadCart(&Cart);
		fatal("Out of memory...");
	}
	printf("[ OK ]\n");
	FenX = 0;
	FenY = 0;
	ppu.MirrorDir = Cart.flags.Mirroring;

	if ( MapperChoice(&Cart) == -1 )
		return;

	memset(Memory,0,0x8000);
	InitPaddle(&P1);	
	if (InitPPU(&ppu) != 0)
	{
		UnloadCart(&Cart);
		free(Memory);
		fatal("PPU Initialisation error..");
	}

	InitMapper(&Cart);
	printf("Initialising XLib !!!");
	if(!XGOn())
	{
		UnloadCart(&Cart);
		free(Memory);
		DestroyPPU();
		fatal("Can't Initialise XLib !...");
	}
	
	Buffer = XNewGPlan();
	XGPlanc(Buffer);
	ScanLine = 0;
	MainCPU.IPeriod = 135; // Do a loop every HBlank

	Reset6502(&MainCPU);

	MainCPU.Trace = 0;

	Run6502(&MainCPU);
	clrscr();
	printf("Freeing memory...\n");
	free(Memory);
	DestroyPPU();
	printf("Emulator is stoping....\n");
	XGOff();
	XDestrGPlan(Buffer);	
	UnloadCart(&Cart);
	LCD_restore(SavedScreen);
	ST_showHelp("TI-NESulator by Godzil");
	exit(0);
}

/** Rd6502()/Wr6502/Op6502() *********************************/
/** These functions are called when access to RAM occurs.   **/
/** They allow to control memory access. Op6502 is the same **/
/** as Rd6502, but used to read *opcodes* only, when many   **/
/** checks can be skipped to make it fast. It is only       **/
/** required if there is a #define FAST_RDOP.               **/
/************************************ TO BE WRITTEN BY USER **/
void Wr6502(register word Addr,register byte Value) /* Write to memory */
{
	unsigned short i;

	if ( MapperWriteHook(Addr,Value) != 0 )
		return;

	if (Addr<0x8000)
	{
		if ((Addr&0xF000) == 0x2000)
		{
			WritePPUReg((char)Addr&0x0F,Value);
		}
		else if (Addr == 0x4014)
		{
			for(i=0;i<256;i++)
				ppu.SPRRAM[i] = Rd6502(Value*0x100+i);
		}
		else if (Addr == 0x4016)
		{
			WritePaddle(&P1,Value);
		}
		else
			Memory[Addr] = Value;
	}
	/* No need for else since after 0x8000 is normaly Read Only (ROM) */
}


byte Rd6502(register word Addr asm("d0")) /* Read memory for normal use */
{ /* In future here need to add support for mapped IO */
	if (Addr<0x8000)
	{
		if ((Addr&0xF000) == 0x2000)
		{
			return ReadPPUReg((char) Addr&0x0F);
		}
		else if (Addr == 0x4016)
		{
			return ReadPaddle(&P1);
		}
		/*else if (Addr == 0x4017)
		{

		}*/
		return Memory[Addr];
	}
	else
	{ /* it is really needed here ? since i'm not sure, i do it.. */
		if (Addr<0xC000)
			return mLBank[Addr & 0x3FFF];
		else
			return mUBank[Addr & 0x3FFF];
	}
}

//byte Op6502(register word Addr asm("d0")) /* Read OpCodes */
//{  /* This func need to be really fast.. but it's not well implemented.. need changes... */
/*	if (Addr<0x8000)
		return Memory[Addr];
	else
	{
		if (Addr<0xC000)
			return mLBank[Addr & 0x3FFF];
		else
			return mUBank[Addr & 0x3FFF];
	}
}*/
asm("
.even
.global Op6502
Op6502:
 cmp.w #0x7FFF, %d0
 bhi next
 move.l Memory, %a0
 and.l #0xFFFF, %d0
 move.b (%a0,%d0),%d0
 rts
 

 .even
next:
 cmp.w #0xBFFF, %d0
 bls next2
 move.l mUBank,%a0
 and.l #0x3FFF, %d0
 move.b (%a0,%d0),%d0
 rts 

  .even
next2:
 move.l mLBank,%a0
 and.l #0x3FFF, %d0
 move.b (%a0,%d0),%d0
 rts
");
/** Loop6502() ***********************************************/
/** 6502 emulation calls this function periodically to      **/
/** check if the system hardware requires any interrupts.   **/
/** This function must return one of following values:      **/
/** INT_NONE, INT_IRQ, INT_NMI, or INT_QUIT to exit the     **/
/** emulation loop.                                         **/
/************************************ TO BE WRITTEN BY USER **/
byte Loop6502(register M6502 *R)
{
	byte ret;
	ret = INT_NONE;
	if (ScanLine == 239)
	{ /* We enter in Vblank time */
		ppu.In_VBlank = 1;
		if (PPUVBlank() == 1)
		{
			ret = INT_NMI;
		}
/*		frame++;
		printf("%d",frame);*/
		ScanLine ++;
	}
	if (ScanLine >= 262)
	{ /* End of VBlank Time */
		ppu.StatusRegister.s.VBlankOccur = 0;
		ppu.StatusRegister.s.Sprite0Occur = 0;
		ppu.In_VBlank = 0;
		ScanLine = 0;
	}
	else
	{ /* Do everithing needed else */
		if ( (ppu.HitSpriteAt < 239) && (ScanLine == ppu.HitSpriteAt) )
		{
			ppu.StatusRegister.s.Sprite0Occur = 1;
		}
		if ( (ppu.ControlRegister2.s.BgVisibility == 1) &&  (ScanLine < 240) )
		{
			ppu.TimedTmpPtr[ScanLine] = ppu.TmpVRamPtr;
			ppu.TimedVScroll[ScanLine] = ppu.VScroll;
			ppu.TimedHScroll[ScanLine] = ppu.HScroll;
		}
		else
		{
			ppu.TimedTmpPtr[0] = ppu.TmpVRamPtr;
			ppu.TimedVScroll[0] = ppu.VScroll;
			ppu.TimedHScroll[0] = ppu.HScroll;
		}
		ScanLine ++;
	}
  XJoyPad();
	if (Press.Esc)
	{
#ifdef DEBUG
		R->Trace = 1;
#else
		return INT_QUIT;
#endif
	}
  if (ScanLine > 240)
  {
		if (Press.B4)
		{
			if (Press.Up)
			{
				FenY += (FenY<(240-Scr.xy.y1))?1:0;
			}
			if(Press.Left)
			{
				FenX += (FenX<(256-Scr.xy.x1))?1:0;
			}
			if(Press.Down)
			{
				FenY -= (FenY>0)?1:0;
			}
			if(Press.Right)
			{
				FenX -= (FenX>0)?1:0;
			}
	  }
	  if (Press.B3)
		{
			if(Press.Left)
			{
				R->IPeriod --;
			}
			if(Press.Right)
			{
				R->IPeriod ++;
			}
	  }
  }
	return ret;
}

/** Patch6502() **********************************************/
/** Emulation calls this function when it encounters an     **/
/** unknown opcode. This can be used to patch the code to   **/
/** emulate BIOS calls, such as disk and tape access. The   **/
/** function should return 1 if the exception was handled,  **/
/** or 0 if the opcode was truly illegal.                   **/
/************************************ TO BE WRITTEN BY USER **/
byte Patch6502(register byte Op,register M6502 *R)
{
	R->Trace = 1;
	printf("Invalid Opcode : 0x%X!\n",Op);
	return 1;
}