/* 
 * TI-NESulator project : mappers.c
 *
 * Generic definition for mappers (use functions pointers)
 *
 * Code Copyright  2003 by TRAPIER Manoel
 *
 */

#include <tigcclib.h>

#include "NESCarts.h"
#include "ppu.h"
#include "M6502.h"
#include "mappers.h"

char MapperName[99][30] = {
"No mapper - Hard wired",
"Nintendo MMC1",
"CNROM",
"UNROM",
"Nintendo MMC3",
"Nintendo MMC5",
"FFE F4xxx",
"AOROM",
"FFE F3xxx",
"Nintendo MMC2",
"Nintendo MMC4",
"ColorDreams chip",
"FFE F6xxx",
"","",
"100-in-1",
"Bandai chip",
"FFE F8xxx",
"Jaleco SS8806 chip",
"Namcot 106 chip",
"Nintendo DiskSystem",
"Konami VRC4a",
"Konami VRC2a",
"Konami VRC2a",
"Konami VRC6",
"Konami VRC4b",
"","","","","","",
"Irem G-101",
"Taito TC0190/TC0350",
"32kB ROM",
"","","","","","","","","","","","","","","","","","","","","","","","","","","","","",
"Tengen RAMBO-1 chip",
"Irem H-3001 chip",
"GNROM",
"SunSoft3 chip",
"SunSoft4 chip",
"SunSoft5 FME-7 chip",
"",
"Camerica chip",
"","","","","","",
"Irem 74HC161/32-based"
};

int (*InitMapper)(NesCart *cart);

int (*MapperWriteHook)(register word Addr,register byte Value);

NesCart *Cart;
extern unsigned char *Memory;
extern unsigned char *mLBank, *mUBank;
extern PPU ppu;

/* Here put include for mappers */

#include "unrom.h"
#include "cnrom.h"
//#include "mmc1.h"

/* ---------------------------- */

int m0_InitMapper(NesCart *cart)
{
	mLBank = cart->ROMBanks[0];
	if (cart->nbROMBanks > 1)
		mUBank = cart->ROMBanks[1];
	else
		mUBank = cart->ROMBanks[0];
	if (cart->nbVROMBanks == 1)
		memcpy(ppu.Memory,cart->VROMBanks[0],0x2000);
	
	return 0;
}

int m0_MapperWriteHook(register word Addr,register byte Value)
{
	return 0;
}


int MapperChoice(NesCart *cart)
{
	Cart = cart;

  printf("Recognized mapper %s\n",MapperName[Cart->Mapper]);

	switch(Cart->Mapper)
	{
	default:
		printf("Unsuported mapper\n");
		return -1;
	case 0:
		InitMapper = m0_InitMapper;
		MapperWriteHook = m0_MapperWriteHook;
		break;
/*	case 1: // MMC 1
		InitMapper = mmc1_InitMapper;
		MapperWriteHook = mmc1_MapperWriteHook;
		break;*/
	case 2: // UNROM
		InitMapper = unrom_InitMapper;
		MapperWriteHook = unrom_MapperWriteHook;
		break;
	case 3: // CNROM
		InitMapper = cnrom_InitMapper;
		MapperWriteHook = cnrom_MapperWriteHook;
		break;
	}
	return 0;
}
