/* 
 * TI-NESulator project : mmc1.h
 *
 * MMC1 mapper
 *
 * Code Copyright © 2003 by TRAPIER Manoel
 *
 */

unsigned char MMC1_reg0;
unsigned char MMC1_reg1;
unsigned char MMC1_reg2;
unsigned char MMC1_reg3;

#define MMC1_R0_MIRROR    0x01
#define MMC1_R0_ONSCREEN  0x02
#define MMC1_R0_PRGAREA   0x04
#define MMC1_R0_PRGSIZE   0x08
#define MMC1_R0_VROMSW    0x10
#define MMC1_R0_RESET     0x80

#define MMC1_R1_VROMB1    0x0F
#define MMC1_R1_256KSEL   0x10
#define MMC1_R1_RESET     0x80

#define MMC1_R2_VROMB2    0x0F
#define MMC1_R2_256KSEL   0x10
#define MMC1_R2_RESET     0x80

#define MMC1_R3_VROMB2    0x0F
#define MMC1_R3_RESET     0x80

int mmc1_InitMapper(NesCart *cart)
{
	MMC1_reg0 = 0;
	MMC1_reg1 = 0;
	MMC1_reg2 = 0;
	MMC1_reg3 = 0;

	mLBank = cart->ROMBanks[0];

	mUBank = cart->ROMBanks[cart->nbROMBanks - 1];
	
	return 0;
}

int mmc1_MapperWriteHook(register word Addr,register byte Value)
{
	static unsigned char Bit = 0;
	static unsigned char LastPort = 5;
	static unsigned char BitBuf = 0;

	unsigned short ROMBankPtr;

	unsigned char *regptr;
	
	if ( Addr > 0x7FFF ) /* Try to write to the rom */
	{
		Addr -= 0x8000;
		if ((Addr/0x2000) != LastPort)
		{
			Bit = 0;
			LastPort = (Addr/0x2000);
			BitBuf = 0;
		}
		switch(Addr/0x2000)
		{
		case 0: /* Reg 0 : 0x8000 > 0x9FFF */
			regptr = &MMC1_reg0;
			break;
		case 1: /* Reg 1 : 0xA000 > 0xBFFF */
			regptr = &MMC1_reg1;
			break;
		case 2: /* Reg 2 : 0xC000 > 0xDFFF */
			regptr = &MMC1_reg2;
			break;
		case 3: /* Reg 3 : 0xE000 > 0xFFFF */
			regptr = &MMC1_reg3;
			break;
		}
		if ( (Value & 0x01) == 1)
		{	
			*regptr |= 1 << Bit;
		}
		
		if ( Value == 0xff ) *regptr = 0;

		if (Bit == 4)
		{ /* Apply */
			switch(LastPort)
			{
			case 0:
				ppu.MirrorDir = 1;
				if ((MMC1_reg0 & MMC1_R0_MIRROR) == 0)
					ppu.MirrorDir = 0;

				ppu.ScreenType = 1;
				if ((MMC1_reg0 & MMC1_R0_ONSCREEN) == 0)
					ppu.MirrorDir = 0;

				break;
			case 1:
				if (Cart->nbVROMBanks > 0)
					if ( (MMC1_reg0 & MMC1_R0_VROMSW) == 0 )
					{ /* 4k Data for PPU 0x0000 */
	printf(".");
					}
					else
					{ /* 8k Data for PPU 0x0000 */
	printf(".");
					}

				break;
			case 2:
				if (Cart->nbVROMBanks > 0)
					if ( (MMC1_reg0 & MMC1_R0_VROMSW) == 0 )
					{ /* 4k Data for PPU 0x1000 */
	printf(".");
					}
				
				break;
			case 3:
				if (Cart->ROMSize >= 512 * 1024) /* 512 and more Ko cart */
				{ 
	printf(".");
				}
				else 
				{ /* normal cart */
					ROMBankPtr = MMC1_reg3 & 0x0F;
					/*
					#define MMC1_R0_PRGAREA   0x04
					#define MMC1_R0_PRGSIZE   0x08
					*/
					if ((MMC1_reg0 & MMC1_R0_PRGSIZE) == 0)
					{ /* 32k Block */
						mLBank = Cart->ROMBanks[ROMBankPtr];
						mUBank = Cart->ROMBanks[ROMBankPtr+1];
					}
					else
					{ /* 16k Block */
						if ( (MMC1_reg0 & MMC1_R0_PRGAREA) == 0 )
						{ /* Write to 0xC000 */
							mUBank = Cart->ROMBanks[ROMBankPtr];
						}
						else
						{ /* Write to 0x8000 */
							mLBank = Cart->ROMBanks[ROMBankPtr];
						}
					}
				}
				break;
			}
		}
		Bit = (Bit == 7)?0:Bit+1;
		return 1;
	}
	return 0;
}
