/* 
 * TI-NESulator project : ppu.c
 *
 * Define and emulate the PPU (Picture Processing Unit) of the real NES
 *
 * Code Copyright  2003 by TRAPIER Manoel
 *
 */
#include <tigcclib.h>

#include "ppu.h"
#include "M6502.h"

#define GetTileColor(tile) ( ( ppu.Memory[tile+y1] & (1<<x1) ) == 0 ? 0 : 1<<1 ) | \
( ( ppu.Memory[tile+y1+8] & (1<<x1) ) == 0 ? 0 : 1 )

#include "X2D.h"

extern PPU ppu;
extern unsigned short ScanLine;
extern void *Buffer;
extern unsigned char FenX,FenY;
extern SCR_RECT Scr;

int InitPPU(PPU *ppu)
{
	
	if ((ppu->Memory = (unsigned char*)malloc(0x4000)) == NULL)
		return -1;
	
	/* Initializing register.. */
	ppu->In_VBlank = 0;
	
	ppu->ControlRegister1.b = 0;
	ppu->ControlRegister2.b = 0;
	ppu->StatusRegister.b = 0;
	ppu->SPR_RAMAddr = 0;
	ppu->VRAMAddrReg1.W = 0;
	ppu->VRAMAddrReg2.W = 0;
	ppu->VRAMAddrMode = 0;
	
	ppu->Bg_Pattern_Table = 0x0000;
	ppu->Name_Table_Addresse = 0x2000;
	ppu->Sprt_Pattern_Table = 0x0000;
	ppu->PPU_Inc = 1;
	
	ppu->MirrorDir = 0;
	ppu->ScreenType = 1;
	return 0;
}

void DestroyPPU()
{
	free(ppu.Memory);
}

PPUSprite PPUGetSprite(unsigned short i)
{
	PPUSprite ret;
	ret.y = ppu.SPRRAM[i*4];
	ret.tileid = ppu.SPRRAM[i*4+1];
	ret.flags.BgPrio = (ppu.SPRRAM[i*4+2]&(1<<5))==0?0:1;
	ret.flags.HFlip = (ppu.SPRRAM[i*4+2]&(1<<6))==0?0:1;
	ret.flags.BgPrio = ppu.SPRRAM[i*4+2]&0x03;
	ret.flags.VFlip = (ppu.SPRRAM[i*4+2]&(1<<7))==0?0:1;
	ret.x = ppu.SPRRAM[i*4+3];
	return ret;
}

unsigned char PPU_Rd(unsigned short Addr)
{
	if ( ppu.ScreenType == 0 )
	{ /* 1 Screen Mode */
		return ppu.Memory[(Addr & 0x00FF)+0x2000];
	}
	else
	{ /* Miroring mode */
		if (ppu.MirrorDir == 0)
		{ /* Horizontal */
			/* Real NT are 0x2000 and 0x2400 */
			if ( Addr >= 0x2C00 )
				return ppu.Memory[Addr - 0xC00];
			else
				return ppu.Memory[Addr];
		}
		else
		{ /* Vertical */
			/* Real NT are 0x2000 and 0x2C00 */
			if (Addr < 0x2C00)
				return ppu.Memory[(Addr&0xFF00)==0x2000?Addr:Addr-0x400];
			else
				return ppu.Memory[(Addr&0xFF00)==0x2C00?Addr:Addr-0x400];
		}
	}
}

void PPU_Wr(unsigned short Addr, unsigned char Value)
{
if ( ppu.ScreenType == 0 )
	{ /* 1 Screen Mode */
		ppu.Memory[(Addr & 0xF0FF)] = Value;
	}
	else
	{ /* Miroring mode */
		if (ppu.MirrorDir == 0)
		{ /* Horizontal */
			/* Real NT are 0x2000 and 0x2C00 */
			if (Addr < 0x2C00)
				ppu.Memory[(Addr & 0xF0FF)] = Value;
			else
				ppu.Memory[(Addr & 0xF0FF) + 0x0C00] = Value;
		}
		else
		{ /* Vertical */
			/* Real NT are 0x2000 and 0x2400 */
			if ( Addr >= 0x2C00 )
				ppu.Memory[Addr - 0xC00] = Value;
			else
				ppu.Memory[Addr] = Value;
		}
	}
}

void PPUDispSprite(unsigned char bg)
{
	int x,y,x1,y1,px,py,i;
	char Color;
	PPUSprite sprite;
	
	short SprtAddr;
	//	int VFlag,HFlag;
	for(i=63;i>=0;i--)
	{
		sprite = PPUGetSprite(i);
		if ( ( sprite.flags.BgPrio == bg ) || ( i == 0 ) )
		{ /* On doit afficher ce sprite */
			y = sprite.y;
			if (y<248)
			{
				SprtAddr = ( (sprite.tileid) << 4 ) + ppu.Sprt_Pattern_Table;
				x = sprite.x;
				for(y1=0;y1<8;y1++)
				{ 
					for(x1=0;x1<8;x1++)
					{
						Color = GetTileColor(SprtAddr);
						if (Color > 0)
						{
							px = x+(sprite.flags.HFlip==0?(7-x1):x1);
							py = y+(sprite.flags.VFlip==0?y1:(7-y1));
							if ( ( i == 0 ) && /* (XGPxlTest(px-FenX,py-FenY) != 0) && */( ppu.HitSpriteAt == 255 ) )
							{
								ppu.HitSpriteAt = py+1;
							}
							XGPxlPut(px-FenX,py-FenY,Color);
						}
					}
				}
			}
		}
	}
}

int PPUVBlank()
{
	int x,y;
	
	unsigned char Color;
	short TileID;
	
	unsigned char XScroll, YScroll;
	
	ppu.StatusRegister.s.VBlankOccur = 1;
	
	XClrGPlan();
	ppu.HitSpriteAt = 255;
	/* A faires les choses qui faut faire durant un lanc de vblank, comme dessiner par ex.. */	
	if (ppu.ControlRegister2.s.SpriteVisibility == 1)
		PPUDispSprite(1);
	
	if (ppu.ControlRegister2.s.BgVisibility == 1)
	{
		/* Display BG with scrolling informations */
		/* J'ai la solution :D */
		/*
		frame start (line 0) (if background and sprites are enabled):
		v=t
		*/
		
		/* Putain de scrolling il y a toujours cette *** de dsynchro.... je commence a me demander si
		   ce n'est pas le core CPU qui est en cause.... 
		 */
		
		ppu.VRAMAddrReg2.W = ppu.TimedTmpPtr[0] + 0x2000;
		YScroll = ppu.TimedHScroll[0];

		for(y=0;y<240;y++)
		{
			if ( (FenY < y ) && ( y < (Scr.xy.y1+FenY)))
      {
			/*
			scanline start (if background and sprites are enabled):
	 		  8421 8421 8421 8421   8421 8421 8421 8421
			v:0000 0100 0001 1111=t:0000 0100 0001 1111
			*/
				ppu.VRAMAddrReg2.W = ( ppu.VRAMAddrReg2.W & ~0x041F ) | ( (ppu.TimedTmpPtr[y]+ 0x2000) & 0x041F );
	
				TileID = ( PPU_Rd(ppu.VRAMAddrReg2.W) << 4 ) + ppu.Bg_Pattern_Table + YScroll;
				XScroll = ppu.TimedHScroll[y];
				
				for(x=0;x<256;x++)
				{
					if ( (FenX < x) && (x < (Scr.xy.x1+FenX)) )
					{
						/*				
						you can think of bits 0,1,2,3,4 of the vram address as the "x scroll"(*8)
						that the ppu increments as it draws.  as it wraps from 31 to 0, bit 10 is
						switched.  you should see how this causes horizontal wrapping between name
						tables (0,1) and (2,3).
						*/
						/* Calculer la couleur du point */
						/* Bits 1 et 0 */
						Color = ( ( ppu.Memory[TileID] & (1<<(7-XScroll)) ) == 0 ? 0 : 1<<1 ) | \
							    ( ( ppu.Memory[TileID+8] & (1<<(7-XScroll)) ) == 0 ? 0 : 1 );
						/* Bits 3 et 2 */
						/*
						XScroll : 0,1,2,3,4
						X = (ppu.VRAMAddrReg2.W & 0x1F)
		
						Y :5,6,7,8,9
						Y = (ppu.VRAMAddrReg2.W & 0x3E0) >> 5
		
						Ptr = 0x23C0;
						Ptr += (ppu.VRAMAddrReg2.W & 0x400);
						Ptr += (ppu.VRAMAddrReg2.W & 0x800);
		
						Ptr + ( X / 4 ) + 8 * ( y / 4 ); // Get the good Attribute Byte
		
						
						*/
						//Color |= (  ) << 2;
		
						if (Color > 0)
							XGPxlPut(x-FenX,y-FenY,Color);
							/*Buffer->line[y][x] = Color;*/
					}							
					XScroll++;
					XScroll &= 7;
					if( XScroll == 0 )
					{ /* On incrmente le compteur de tile :| */
						if ( (ppu.VRAMAddrReg2.W & 0x1F) == 0x1F )
						{ /* On met a 0 et change l'etat du bit 10 */
							ppu.VRAMAddrReg2.W &= ~0x1F;
							
							/* A voir si ya pas bcp mieux */
							if ( (ppu.VRAMAddrReg2.W & 0x400) == 0 )
								ppu.VRAMAddrReg2.W |= 0x400;
							else
								ppu.VRAMAddrReg2.W &= ~0x400;
						}
						else
						{ /* on incremente juste */
							ppu.VRAMAddrReg2.W = (ppu.VRAMAddrReg2.W & ~0x1F) | ( ( ppu.VRAMAddrReg2.W & 0x1F ) + 1);
						}
						//TileID = ( ppu.Memory[ppu.VRAMAddrReg2.W] << 4 ) + ppu.Bg_Pattern_Table + YScroll;
						TileID = ( PPU_Rd(ppu.VRAMAddrReg2.W) << 4 ) + ppu.Bg_Pattern_Table + YScroll;
					}
				}
				/*
				you can think of bits 5,6,7,8,9 as the "y scroll"(*8).  this functions
				slightly different from the X.  it wraps to 0 and bit 11 is switched when
				it's incremented from _29_ instead of 31.  there are some odd side effects
				from this.. if you manually set the value above 29 (from either 2005 or
				2006), the wrapping from 29 obviously won't happen, and attrib data will be
				used as name table data.  the "y scroll" still wraps to 0 from 31, but
				without switching bit 11.  this explains why writing 240+ to 'Y' in 2005
				appeared as a negative scroll value.
				*/
			}
			YScroll++;
			YScroll &= 7;
			if( YScroll == 0 )
			{ /* On incrmente le compteur de tile :| */
				if ( (ppu.VRAMAddrReg2.W & 0x3E0) == 0xEA0 )
				{ /* On met a 0 et change l'etat du bit 10 */
					ppu.VRAMAddrReg2.W &= ~0x3F0;
					
					/* A voir si ya pas bcp mieux */
					if ( (ppu.VRAMAddrReg2.W & 0x800) == 0 )
						ppu.VRAMAddrReg2.W |= 0x800;
					else
						ppu.VRAMAddrReg2.W &= ~0x800;
				}
				else
				{ /* on incremente juste */
					ppu.VRAMAddrReg2.W = (ppu.VRAMAddrReg2.W & ~0x3F0) | ( ( ( (ppu.VRAMAddrReg2.W & 0x3F0) >> 5 ) + 1) << 5 );
				}
			}
		}
	}
	
	if (ppu.ControlRegister2.s.SpriteVisibility == 1)
		PPUDispSprite(0);
	
	
	XCpyGPlanToLCD();
	
	if (ppu.ControlRegister1.s.VBlank_NMI == 1)
		return 1;
	
	return 0;
}

inline char ReadPPUReg(char RegID asm("d1"))
{ 
	/* RegID is the nb of the reg 0-7 */
	unsigned char ret;
	switch(RegID)
	{
	default:
		ret = (unsigned char) 0xCD; /* Can return every thing you want here :D */
		break;
	case 2:
		ret = ppu.StatusRegister.b;
		ppu.StatusRegister.s.VBlankOccur = 0;
		ppu.StatusRegister.s.Sprite0Occur = 0;
		ppu.VRAMAddrMode = 0;
		break;
	case 7: /* PPU in NES is really strange.. Bufferised send is nerd.. */
		if ( ppu.VRAMAddrReg2.W < 0x3EFF)
		{
			ret = ppu.VRAMBuffer;
			ppu.VRAMBuffer = ppu.Memory[ppu.VRAMAddrReg2.W];
			ppu.VRAMAddrReg2.W += ppu.PPU_Inc;
		}
		else
			ret = ppu.Memory[ppu.VRAMAddrReg2.W];
		break;
	}
	return ret;
}

inline void WritePPUReg(char RegID, unsigned char val)
{ 
	/* RegID is the nb of the reg 0-7 */
	
	/*
	PPUCC = CPUCC * 3
	Scanline = PPUCC div 341 - 21
	PixelOfs = (PPUCC+16) mod 341
	CPUcollisionCC = ((Y+21)*341+X)/3
	*/
	
	switch(RegID)
	{
	default: /* For not writeable reg */
		break;
	case 0: /* Control Register 1 */
		ppu.ControlRegister1.b = val;
		
		ppu.Bg_Pattern_Table = (ppu.ControlRegister1.s.BgPattern == 1)?0x1000:0x0000;
		ppu.Sprt_Pattern_Table = (ppu.ControlRegister1.s.SptPattern == 1)?0x1000:0x0000;
		ppu.PPU_Inc = (ppu.ControlRegister1.s.AddrIncrmt == 1)?32:1;
		
		ppu.Name_Table_Addresse = 0x2000;
		switch(ppu.ControlRegister1.s.NameTblAddr)
		{
		case 3:
			ppu.Name_Table_Addresse += 0x400;
		case 2:
			ppu.Name_Table_Addresse += 0x400;
		case 1:
			ppu.Name_Table_Addresse += 0x400;
		case 0:
			break;
		}
		
		ppu.TmpVRamPtr = ( ( ppu.TmpVRamPtr & ~0xC00 ) | (ppu.ControlRegister1.s.NameTblAddr << 10) );

		break;
		case 1: /* Control Register 2 */
			ppu.ControlRegister2.b = val;
			break;
		case 3: /* SPR-RAM Addresse Register */
			ppu.SPR_RAMAddr = val;
			break;
		case 4: /* SPR-RAM Input Register */
			ppu.SPRRAM[ppu.SPR_RAMAddr] = val;
			break;
		case 5: /* VRAM Address register 1 */
			if (ppu.VRAMAddrMode == 0)
			{
			/*
			2005 first write:
			t:0000000000011111=d:11111000
			x=d:00000111
				*/
				/*ppu.VRAMAddrReg1.B.h = val;*/
				ppu.VRAMAddrMode = 1;
				
				ppu.TmpVRamPtr = ( ( ppu.TmpVRamPtr & 0x1F ) | (val>>3) );
				ppu.HScroll = val & 0x7;
			}
			else
			{
			/*
			2005 second write:
			t:0000001111100000=d:11111000
			t:0111000000000000=d:00000111
				*/
				/*ppu.VRAMAddrReg1.B.l = val;
				ppu.VRAMAddrReg1.W &= 0x3FFF;*/
				ppu.VRAMAddrMode = 0;
				ppu.VScroll = 0;
				ppu.TmpVRamPtr = ( ( ppu.TmpVRamPtr & ~0x03E0 ) | ((val&0xF8)<<2) );
				ppu.TmpVRamPtr = ( ( ppu.TmpVRamPtr & ~0x7000 ) | ((val&0x07)<<12) );
		//		printf("/0x%04X_",ppu.TmpVRamPtr);
				/*if (ScanLine<240)*/
//					printf("%d -> 2005: 0x%04X\n",ScanLine,ppu.TmpVRamPtr);
			}
			break;
		case 6: /* VRAM Address register 2 */
			if (ppu.VRAMAddrMode == 0)
			{
				/*ppu.VRAMAddrReg2.B.h = val;*/
				ppu.VRAMAddrMode = 1;
				/*
				2006 first write:
				t:0011111100000000=d:00111111
				t:1100000000000000=0
				*/
				ppu.TmpVRamPtr &= ~0x3F00;
				ppu.TmpVRamPtr |= (((unsigned long)val & 0x3F)<<8);
				ppu.TmpVRamPtr = ( ppu.TmpVRamPtr & 0x3FFF );
			}
			else
			{
			/*ppu.VRAMAddrReg2.B.l = val;
				ppu.VRAMAddrReg2.W &= 0x3FFF;*/
				ppu.VRAMAddrMode = 0;
				/*
				2006 second write:
				t:0000000011111111=d:11111111
				v=t
				*/
				ppu.TmpVRamPtr = ( ( ppu.TmpVRamPtr & ~0x00FF ) | ( val ) );
				ppu.VRAMAddrReg2.W = ppu.TmpVRamPtr;			
				/*if (ScanLine<240)
					printf("%d -> 2006: 0x%04X\n",ScanLine,ppu.TmpVRamPtr);*/
			}
			break;
		case 7: /* VRAM I/O */
			ppu.Memory[ppu.VRAMAddrReg2.W] = val;
			ppu.VRAMAddrReg2.W += ppu.PPU_Inc;
			break;
	}
}