/* 
 * TI-NESulator project : ppu.h
 *
 * Define and emulate the PPU (Picture Processing Unit) of the real NES
 *
 * Code Copyright  2003 by TRAPIER Manoel
 *
 */

/* PPU Registers */

/* 
   (R-) mean Read Only
   (-W) mean Write Only
   (RW) mean ReadWrite
*/

#include "M6502.h"

#ifndef PPU_H
#define PPU_H

typedef struct PPU_
{
	union uCtrReg1_ {
		struct s1_ {
#ifdef LSB_FIRST
			unsigned int NameTblAddr :2 ; /* 00 : VRAM 0x2000
                                     01 : VRAM 0x2400
                                     10 : VRAM 0x2800
                                     11 : VRAM 0x2C00 */
			unsigned int AddrIncrmt  :1 ; /* 0 : Inc by 1
			                         1 : Inc by 32 */
			unsigned int SptPattern  :1 ; /* 0 : VRAM 0x0000
                    	                 1 : ppu 0x1000 */
			unsigned int BgPattern   :1 ; /* 0 : VRAM 0x0000
	                                     1 : ppu 0x1000 */
			unsigned int SpriteSize  :1 ; /* 0 : 8x8
  	                                   1 : 16x16 */
			unsigned int Unused      :1; /* Normaly determine Master or Slave, but unused on NES */
			unsigned int VBlank_NMI  :1 ; /* 0 : NMI Disable  
    	                                 1 : NMI Enable */			
#else
			unsigned int VBlank_NMI  :1 ; /* 0 : NMI Disable  
      	                               1 : NMI Enable */			
			unsigned int Unused      :1; /* Normaly determine Master or Slave, but unused on NES */
			unsigned int SpriteSize  :1 ; /* 0 : 8x8
        	                             1 : 16x16 */
			unsigned int BgPattern   :1 ; /* 0 : VRAM 0x0000
          	                           1 : ppu 0x1000 */                                     			                                     
			unsigned int SptPattern  :1 ; /* 0 : VRAM 0x0000
            	                         1 : ppu 0x1000 */
			unsigned int AddrIncrmt  :1 ; /* 0 : Inc by 1
        			                         1 : Inc by 32 */                                     
			unsigned int NameTblAddr :2 ; /* 00 : VRAM 0x2000
              	                       01 : VRAM 0x2400
               		                     10 : VRAM 0x2800
	                                     11 : VRAM 0x2C00 */
#endif                                     
		} s;
		unsigned char b;
	} ControlRegister1; /* CPU RAM 0x2000 (-W) */
	
	union uCtrReg2_ {
		struct s2_ {
#ifdef LSB_FIRST
			unsigned int DisplayType      :1;/* 0 : Color display
																			    1 : Monochrome display */
			unsigned int BgClipping       :1;/* 0 : Left 8pixel BG is Invisible
																			    1 : No clipping to background */
			unsigned int SpriteClipping   :1;/* 0 : Left 8pixel Sprite are invisibles
																			    1 : Sprite aren't clipped */
			unsigned int BgVisibility     :1;/* 0 : Background not displayed
																			    1 : Background is show */
			unsigned int SpriteVisibility :1;/* 0 : Sprites aren't diplayed
																					1 : Sprite are displayed */
			unsigned int Colour           :3 ;/* Used as BgColor when DisplayType is Colour
																			     or used as Colour intensify where Monochrome
																	 	        000 : None
																		        001 : Green
																		        010 : Blue
																		        100 : Red */
#else								       
			unsigned int Colour           :3 ;/* Used as BgColor when DisplayType is Colour
																			     or used as Colour intensify where Monochrome
							 	    										    000 : None
																		        001 : Green
																		        010 : Blue
																		        100 : Red */
 			unsigned int SpriteVisibility :1;/* 0 : Sprites aren't diplayed
																					1 : Sprite are displayed */
			unsigned int BgVisibility     :1;/* 0 : Background not displayed
																			    1 : Background is show */
			unsigned int SpriteClipping   :1;/* 0 : Left 8pixel Sprite are invisibles
																			    1 : Sprite aren't clipped */
			unsigned int BgClipping       :1;/* 0 : Left 8pixel BG is Invisible
									  										  1 : No clipping to background */
			unsigned int DisplayType      :1;/* 0 : Color display
																			    1 : Monochrome display */
#endif								      
		} s;
		unsigned char b;
	} ControlRegister2; /* CPU RAM 0x2001 (-W) */

	union uStatReg_ {
		struct s3_ {
#ifdef LSB_FIRST
			unsigned int unused       :4 ;
			unsigned int VRAMProtect  :1 ; /* 0 : VRAM is Writeable 
								  									    1 : VRAM is not writeable */
			unsigned int SprtCount    :1 ; /* 0 : Eight Sprite or Less on current line 
																	      1 : More than 8 sprite on current line */
			unsigned int Sprite0Occur :1 ; /* 0 : Sprite 0 not found
																	      1 : PPU has hit Sprite 0 */
			unsigned int VBlankOccur  :1 ; /* 0 : Not in VBlank
																	      1 : VBlank */
#else
			unsigned int VBlankOccur  :1 ; /* 0 : Not in VBlank
																	      1 : VBlank */
			unsigned int Sprite0Occur :1 ; /* 0 : Sprite 0 not found
																	      1 : PPU has hit Sprite 0 */			
			unsigned int SprtCount    :1 ; /* 0 : Eight Sprite or Less on current line 
								  									    1 : More than 8 sprite on current line */
			unsigned int VRAMProtect  :1 ; /* 0 : VRAM is Writeable 
																	      1 : VRAM is not writeable */								      								      
			unsigned int unused       :4 ;
#endif
		} s;
		unsigned char b;
	} StatusRegister; /* CPU RAM 0x2002 (R-) Note : VBLankOccur is reset when read */ 

	unsigned char SPR_RAMAddr; /* CPU RAM 0x2003 (-W2) */
	
	pair VRAMAddrReg1; /* CPU RAM 0x2005 (-W2) */
	pair VRAMAddrReg2; /* CPU RAM 0x2006 (-W2) */

	char VRAMAddrMode;

	unsigned char VRAMBuffer;

	unsigned char *Memory; /* 32kB */

	unsigned char SPRRAM[256];

	unsigned short Name_Table_Addresse;
	unsigned short Bg_Pattern_Table;
	unsigned short Sprt_Pattern_Table;
	unsigned char  PPU_Inc;

	unsigned short TimedTmpPtr[240];
	unsigned char TimedVScroll[240];
	unsigned char TimedHScroll[240];

	unsigned short TmpVRamPtr;

	unsigned char HitSpriteAt;

	unsigned char In_VBlank;
	unsigned char VScroll;
	unsigned char HScroll;

	unsigned char ScreenType;/* 0 : Single Screen at $2000
															1 : Normal mode */

	unsigned char MirrorDir; /* 0 : Horizontal
						 									1 : Vertical */

} PPU;

typedef struct PPUSprite_
{
	unsigned char y;
	unsigned char tileid;
	struct SprtFlag_
	{
#ifdef LSB_FIRST
		unsigned int UpperColor :2; /* ? */
		unsigned int unuse      :3;
		unsigned int BgPrio     :1; /* 0 : In front of BG
		                               1 : Behind BG */
		unsigned int HFlip      :1; /* 0 : Not Flipped 
		                               1 : Flipped */
		unsigned int VFlip      :1; /* 0 : Not Flipped 
		                               1 : Flipped */
#else
		unsigned int VFlip      :1; /* 0 : Not Flipped 
		                               1 : Flipped */
		unsigned int HFlip      :1; /* 0 : Not Flipped 
		                               1 : Flipped */
		unsigned int BgPrio     :1; /* 0 : In front of BG
		                               1 : Behind BG */		                               
		unsigned int unuse      :3;
		unsigned int UpperColor :2; /* ? */		                               
#endif		                               
	} flags;
	unsigned char x;
} PPUSprite;

/*
PPU must be initialized after memory initialisation..
*/
int InitPPU(PPU *ppu);

int PPUVBlank();

inline char ReadPPUReg(char RegID asm("d1"));

inline void WritePPUReg(char RegID, unsigned char val);

void DestroyPPU();

#endif