/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                       NESCarts.c                        **/
/**                                                         **/
/** This file contains functions to process NES cartridges  **/
/** in the .NES file format. Also see NESCarts.h.           **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1998-2001                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#define MIN_AMS 100           // Compile for AMS 1.00 or higher

#include "NESCarts.h"

#include <tigcclib.h>
//#include <stdlib.h>
//#include <stdio.h>

/** NES_CRC() ************************************************/
/** Calculate 16bit CRC of a given number of bytes.         **/
/*************************************************************/
unsigned short NES_CRC(unsigned char *Data,int N)
{
  unsigned short R;
  int J;

  for(J=0,R=0;J<N;J++) R+=*Data++;
  return(R&0xFFFF);
}


int LoadCart(const char *filename, NesCart *cart)
{
	FILE *fp;
	int i;
	unsigned char *GameCart;

	if ((fp = fopen(filename,"rb")) == NULL)
		return -1;
	
	/* TI Related stuff, very uggly, and need to be changed.. */
	HeapLock(fp->handle);
	GameCart = 2 + HeapDeref(fp->handle);

/* Not used apparely with TIs :(
	fseek(fp,0,SEEK_END);
	filesize = ftell(fp);
	fseek(fp,0,SEEK_SET);


	if ((GameCart = (unsigned char *) malloc(filesize)) == NULL)
		return -2;
	tmp = GameCart;


	while(!feof(fp))
		fread(tmp++,1,1,fp);
	
	fclose(fp);
printf("Readed %d bytes\n",tmp-GameCart);
		*/

	cart->dotNESfile = GameCart;

	cart->nbROMBanks = NES_ROMBanks(GameCart);
	cart->nbVROMBanks = NES_VROMBanks(GameCart);
	cart->ROMSize = NES_ROMSize(GameCart);
	cart->VROMSize = NES_VROMSize(GameCart);
	cart->Mapper = NES_Mapper(GameCart);
	cart->flags.Mirroring = NES_Mirroring(GameCart)==0?0:1;
	cart->flags.Battery = NES_Battery(GameCart)==0?0:1;
	cart->flags.Trainer = NES_Trainer(GameCart)==0?0:1;
	cart->flags.FourScreens = NES_4Screens(GameCart)==0?0:1;
	cart->flags.VSSystem = NES_VSSystem(GameCart)==0?0:1;
	cart->nbRAMBanks = NES_RAMBanks(GameCart);
/*
16-...   ROM banks, in ascending order. If a trainer is present, its
         512 bytes precede the ROM bank contents.
...-EOF  VROM banks, in ascending order.
*/

	cart->dotNESfile = GameCart;
	cart->Trainer = NULL;
	if (cart->flags.Trainer == 1)
		cart->Trainer = GameCart+16;
	
	for(i=0;i<MAX_BANKS;i++)
	{
		cart->ROMBanks[i] = NULL;
		cart->VROMBanks[i] = NULL;
	}
	
	for(i=0;i<cart->nbROMBanks;i++)
	{
		cart->ROMBanks[i] = GameCart+16+(i*1<<14);
	}
	for(i=0;i<cart->nbVROMBanks;i++)
		cart->VROMBanks[i] = GameCart+16+(i*1<<13);
	
	printf("Press a key to continue.\n"); ngetchx();

	printf("Nes cart %s loaded\n",filename);
	printf("Information :\n" \
		"		Number of ROM Banks  : %d\n		Number of VROM Banks : %d\n" \
		"		Number of RAM Banks  : %d\n" \
		"		Total ROM Size       : %ul\n		Total VROM Size      : %ul\n" \
		"		Mapper type          : %d\n		Mirroring ?          : %s\n" \
		"		Battery ?            : %s\n		Trainer ?            : %s\n" \
		"		4VScreen ?           : %s\n		VSSystem ?           : %s\n", \
		cart->nbROMBanks, \
		cart->nbVROMBanks, \
		cart->nbRAMBanks, \
		cart->ROMSize, \
		cart->VROMSize, \
		cart->Mapper, \
		cart->flags.Mirroring==0?"Horizontal":"Vertical", \
		cart->flags.Battery==0?"No":"Yes", \
		cart->flags.Trainer==0?"No":"Yes", \
		cart->flags.FourScreens==0?"No":"Yes", \
		cart->flags.VSSystem==0?"No":"Yes");
	return 0;
}
