/* 
 * TI-NESulator project : main.c
 *
 *
 * Code Copyright  2003 by TRAPIER Manoel
 *
 */

#define USE_TI89              // Compile for TI-89
#define USE_TI92PLUS          // Compile for TI-92 Plus
//#define USE_V200              // Compile for V200

// #define OPTIMIZE_ROM_CALLS // Use ROM Call Optimization

#define MIN_AMS 100           // Compile for AMS 1.00 or higher

#define SAVE_SCREEN           // Save/Restore LCD Contents

#include <tigcclib.h>

#include "M6502.h"

#include "NESCarts.h"

#define V_MAJOR 0
#define V_MINOR 0
#define V_OTHER "prealpha"

unsigned char *Memory;
unsigned char *mLBank, *mUBank;

#define pause() { printf("Press a key to continue.\n"); ngetchx(); }
#define fatal(s) { printf("%s",s);	pause(); ST_showHelp("TI-NESulator by Godzil");	exit(0); }


void _main()
{
	M6502 MainCPU;
	NesCart Cart;

/* Here we will fill the memory */
/*
--------------------------------------- $10000
 Upper Bank of Cartridge ROM
--------------------------------------- $C000
 Lower Bank of Cartridge ROM
--------------------------------------- $8000
 Cartridge RAM (may be battery-backed)
--------------------------------------- $6000
 Expansion Modules
--------------------------------------- $5000
 Input/Output
--------------------------------------- $2000
 2kB Internal RAM, mirrored 4 times
--------------------------------------- $0000
*/
	clrscr();
	FontSetSys(F_4x6);
	printf("TI-NESulator v%d.%d %s by Godzil\nCopyright 2003 TRAPIER Manol\n",V_MAJOR,V_MINOR,V_OTHER);
	printf("Please Wait while loading cartridge...\n");
	if(LoadCart("smb1",&Cart) != 0)
		fatal("Loading error : can't find file !...\n");
	
	pause();
	
	printf("Creating NES memory\n");
	printf("->Non ROM memory (32kB)      ");
	if ( (Memory = (unsigned char *)malloc(0x8000)) == NULL)
		fatal("[ERROR]\n");
	printf("[ OK ]\n");
	
	mLBank = Cart.ROMBanks[0];
	mUBank = Cart.ROMBanks[1];

	pause();

	printf("Reseting CPU...\n");
	MainCPU.IPeriod = 10000;
	Reset6502(&MainCPU);
	MainCPU.Trace = 1;
	printf("Power Up CPU :D\n\n");
	Run6502(&MainCPU);
	
	free(Memory);
	printf("Exiting...");
	return;
}

/** Rd6502()/Wr6502/Op6502() *********************************/
/** These functions are called when access to RAM occurs.   **/
/** They allow to control memory access. Op6502 is the same **/
/** as Rd6502, but used to read *opcodes* only, when many   **/
/** checks can be skipped to make it fast. It is only       **/
/** required if there is a #define FAST_RDOP.               **/
/************************************ TO BE WRITTEN BY USER **/
void Wr6502(register word Addr,register byte Value) /* Write to memory */
{
	if (Addr<0x8000)
		Memory[Addr] = Value;
	/* No need for else since after 0x8000 is normaly Read Only (ROM) */
}
byte Rd6502(register word Addr) /* Read memory for normal use */
{ /* In future here need to add support for mapped IO */
	if (Addr<0x8000)
		return Memory[Addr];
	else
	{ /* it is really neede here ? since i'm not sure, i do it.. */
		if (Addr<0xC000)
			return mLBank[Addr & 0x3FFF];
		else
			return mUBank[Addr & 0x3FFF];
	}
}
byte Op6502(register word Addr) /* Read OpCodes */
{ /* This func need to be really fast.. but it's not well implemented.. need changes... */
	if (Addr<0x8000)
		return Memory[Addr];
	else
	{
		if (Addr<0xC000)
			return mLBank[Addr & 0x3FFF];
		else
			return mUBank[Addr & 0x3FFF];
	}
}

/** Loop6502() ***********************************************/
/** 6502 emulation calls this function periodically to      **/
/** check if the system hardware requires any interrupts.   **/
/** This function must return one of following values:      **/
/** INT_NONE, INT_IRQ, INT_NMI, or INT_QUIT to exit the     **/
/** emulation loop.                                         **/
/************************************ TO BE WRITTEN BY USER **/
byte Loop6502(register M6502 *R)
{
/*	printf("Loop\n");*/
  puts("VBlank\n");
  
	return INT_NONE;
}

/** Patch6502() **********************************************/
/** Emulation calls this function when it encounters an     **/
/** unknown opcode. This can be used to patch the code to   **/
/** emulate BIOS calls, such as disk and tape access. The   **/
/** function should return 1 if the exception was handled,  **/
/** or 0 if the opcode was truly illegal.                   **/
/************************************ TO BE WRITTEN BY USER **/
byte Patch6502(register byte Op,register M6502 *R)
{
	printf("Found Invalid Opcode : 0x%X!\n",Op);
	return 1;
}